/*  glmix - Mix multiple inputs to stereo output with OpenGL visualisation
    Copyright (C) 2003  Nick Lamb <njl195@zepler.org.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* Some of this code is based on public domain examples from gtkglext */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include <jack/jack.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtkgl.h>

#include <GL/gl.h>
#include <GL/glu.h>

#define DEGREES (M_PI / 180.0)

/* Arbitrary non-zero constant */
#define BASE_PORT_NAME 64

#define TIMEOUT_FAST 20
#define TIMEOUT_SLOW 100
#define MOUSE_VELOCITY_FACTOR 600.0

/* Window size */
#define DEFAULT_WIDTH  600
#define DEFAULT_HEIGHT 400

#define WINDOW_TITLE "GL Mixer"
#define APP_NAME "glmix"

#define INPUT_PORTS 8
#define HITLIST_SIZE (INPUT_PORTS + 1) * 4

static char jack_client_name[] = "glmix";
static float jack_level[INPUT_PORTS], draw_level[INPUT_PORTS];
static float jack_vol[INPUT_PORTS] = {1.0, 1.0, 0.4, 0.4, 0.2, 0.2, 0.1, 0.1 };
static float jack_pan[INPUT_PORTS] = {0.0, 1.0, 0.2, 0.8, 0.1, 0.9, 0.2, 0.8 };

static jack_client_t *jack_handle;
static jack_port_t *left_port, *right_port, *input_port[INPUT_PORTS];

static GLuint draw_order[INPUT_PORTS];
static GLdouble zNear = 1.0, zFar = 99.1;
static GLfloat sphi = 0.0, stheta = -15.0;
static GLfloat actual_width, actual_height;
static GLfloat zoom_distance = 11.3;
static int lastX, lastY;
static int selected = 0;

static guint32 redraw_rate = TIMEOUT_SLOW;
static gboolean spin = FALSE;

static void         timeout_speed     (GtkWidget   *widget, guint32 interval);
static void         timeout_none      (GtkWidget   *widget);

static void         toggle_spin       (GtkWidget   *widget);

static GdkGLConfig *configure_gl      (void);

static GtkWidget   *create_popup_menu (GtkWidget   *drawing_area);
static GtkWidget   *create_window     (GdkGLConfig *glconfig);


static int playback_callback(jack_nframes_t nframes, void *arg) {
    int k, frame;

    jack_default_audio_sample_t *left = (jack_default_audio_sample_t *)jack_port_get_buffer(left_port, nframes);
    jack_default_audio_sample_t *right = (jack_default_audio_sample_t *)jack_port_get_buffer(right_port, nframes);
    memset(left, 0, nframes * sizeof(jack_default_audio_sample_t));
    memset(right, 0, nframes * sizeof(jack_default_audio_sample_t));

    for (k= 0; k < INPUT_PORTS; ++k) {
      const float pan = jack_pan[k] * 0.5 * M_PI, vol = jack_vol[k];
      float level = jack_level[k];
      jack_default_audio_sample_t *input = (jack_default_audio_sample_t *)jack_port_get_buffer(input_port[k], nframes);

      for (frame= 0; frame < nframes; ++frame) {
        right[frame] += input[frame] * sin(pan) * vol;
        left[frame] += input[frame] * cos(pan) * vol;
        if (fabsf(input[frame]) > level) level = fabsf(input[frame]);
      }

      jack_level[k] = level;
    }

    return(0);
}

static jack_client_t *open_jack(void) {
    jack_client_t *jack_handle;
            
    if ((jack_handle = jack_client_new(jack_client_name)) == 0) {
        fprintf (stderr, "jack server not running ?\n");
        exit (1);
    }
    input_port[0] = jack_port_register(jack_handle, "input_1", JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
    input_port[1] = jack_port_register(jack_handle, "input_2", JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
    input_port[2] = jack_port_register(jack_handle, "input_3", JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
    input_port[3] = jack_port_register(jack_handle, "input_4", JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
    input_port[4] = jack_port_register(jack_handle, "input_5", JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
    input_port[5] = jack_port_register(jack_handle, "input_6", JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
    input_port[6] = jack_port_register(jack_handle, "input_7", JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
    input_port[7] = jack_port_register(jack_handle, "input_8", JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
    left_port = jack_port_register(jack_handle, "left", JACK_DEFAULT_AUDIO_TYPE,
 JackPortIsOutput, 0);
    right_port = jack_port_register(jack_handle, "right", JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
    jack_set_process_callback(jack_handle, playback_callback, 0);
    return(jack_handle);
}

/* -- ought to initialise OpenGL state here */

static void realize (GtkWidget *widget, gpointer data)
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);

  if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext))
    return;

  gdk_gl_drawable_gl_end (gldrawable);
}

static gboolean configure_event (GtkWidget *widget, GdkEventConfigure *event,
                                 gpointer data)
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);

  actual_width = widget->allocation.width;
  actual_height = widget->allocation.height;

  if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext))
    return FALSE;

  /* make the GL view full the available area */
  glViewport (0, 0, actual_width, actual_height);

  gdk_gl_drawable_gl_end (gldrawable);

  return TRUE;
}

/* How many squares to use in the grid, must be even */
static const float grid = 16.0f;

static void draw_grid(void)
{
  int k;
  GLfloat z = -0.5f;

  glBegin(GL_LINE_STRIP);
  for (k= 0; k <= grid; ++k) {
    glVertex3f( -0.5f + k / grid, 0.0, z);
    z = -z;
    glVertex3f( -0.5f + k / grid, 0.0, z);
  }
  for (k= 0; k <= grid; ++k) {
    glVertex3f( z, 0.0, 0.5f - k / grid);
    z = -z;
    glVertex3f( z, 0.0, 0.5f - k / grid);
  }
  glEnd();
}

static void draw_box(void)
{
  glBegin(GL_TRIANGLE_STRIP);
    glVertex3f(-0.5, 0.0, -0.5);
    glVertex3f(0.5, 0.0, -0.5);
    glVertex3f(-0.5, 1.0, -0.5);
    glVertex3f(0.5, 1.0, -0.5);
    glVertex3f(-0.5, 1.0, 0.5);
    glVertex3f(0.5, 1.0, 0.5);
    glVertex3f(-0.5, 0.0, 0.5);
    glVertex3f(0.5, 0.0, 0.5);
    glVertex3f(-0.5, 0.0, -0.5);
    glVertex3f(0.5, 0.0, -0.5);
  glEnd();

  glBegin(GL_TRIANGLE_STRIP);
    glVertex3f(-0.5, 0.0, -0.5);
    glVertex3f(-0.5, 0.0, 0.5);
    glVertex3f(-0.5, 1.0, -0.5);
    glVertex3f(-0.5, 1.0, 0.5);
  glEnd();
  glBegin(GL_TRIANGLE_STRIP);
    glVertex3f(0.5, 0.0, -0.5);
    glVertex3f(0.5, 0.0, 0.5);
    glVertex3f(0.5, 1.0, -0.5);
    glVertex3f(0.5, 1.0, 0.5);
  glEnd();

  /* outline */
  glBegin(GL_LINE_STRIP);
    glVertex3f(-0.5, 0.0, -0.5);
    glVertex3f(0.5, 0.0, -0.5);
    glVertex3f(0.5, 0.0, 0.5);
    glVertex3f(-0.5, 0.0, 0.5);
    glVertex3f(-0.5, 0.0, -0.5);
    glVertex3f(-0.5, 1.0, -0.5);
    glVertex3f(-0.5, 1.0, 0.5);
    glVertex3f(-0.5, 0.0, 0.5);
  glEnd();

  glBegin(GL_LINE_STRIP);
    glVertex3f(0.5, 1.0, 0.5);
    glVertex3f(-0.5, 1.0, 0.5);
    glVertex3f(-0.5, 1.0, -0.5);
    glVertex3f(0.5, 1.0, -0.5);
    glVertex3f(0.5, 1.0, 0.5);
    glVertex3f(0.5, 0.0, 0.5);
    glVertex3f(0.5, 0.0, -0.5);
    glVertex3f(0.5, 1.0, -0.5);
  glEnd();
}

static void draw_signal(int port)
{
  glPushMatrix();
  glTranslatef (jack_pan[port] * 10.0, 0.0, 10.0 * jack_vol[port]);
  switch (port % 4) {
    case 0:
      glColor4f(0.1, 0.1, 1.0, 0.4); break;
    case 1:
      glColor4f(0.1, 1.0, 0.1, 0.4); break;
    case 2:
      glColor4f(0.1, 1.0, 1.0, 0.4); break;
    case 3:
      glColor4f(1.0, 1.0, 0.1, 0.4); break;
  }
  glScalef(0.5, 2.5, 0.5);
  draw_box();
  if (draw_level[port] > 0.0) {
    glColor4f(0.8, 0.0, 0.0, 1.0);
    glScalef(0.4, draw_level[port], 0.4);
    draw_box();
  }
  glPopMatrix();
}

static void draw_scene(void)
{
  int k;

  glInitNames();
  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();
  glTranslatef (0.0, 0.0, -zoom_distance);
  glRotatef (-stheta, 1.0, 0.0, 0.0);
  glRotatef (sphi, 0.0, 1.0, 0.0);

  /* Draw grid */
  glPushName(0);
  glPushMatrix();
  glColor4f(1.0, 1.0, 1.0, 1.0);
  glScalef(10.5, 1.0, 10.5);
  draw_grid();
  glPopMatrix();

  /* Draw signal boxes */
  glTranslatef (-5.0, 0.0, -5.0);
  for (k = 0; k < INPUT_PORTS; ++k) {
    const int port = draw_order[k];
    glLoadName(BASE_PORT_NAME + port);
    draw_signal(port);
  }
}

static void order_objects(void)
{
  GLuint hitlist[HITLIST_SIZE];
  GLuint depth[INPUT_PORTS];
  GLint hits, k= 0, j= 0;

  for (k= 0; k < INPUT_PORTS; ++k) {
    draw_order[k]= k;
  }
  glSelectBuffer(HITLIST_SIZE, hitlist);
  glRenderMode(GL_SELECT);
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluPerspective (64.0, 1.25, zNear, zFar);
  draw_scene();

  /* Now do the hard part */

  hits= glRenderMode(GL_RENDER);

  for (k= 0; k < hits; ++k) {
    const GLuint name = hitlist[4 * k + 3];
    if (name >= BASE_PORT_NAME)
      depth[name - BASE_PORT_NAME]= hitlist[4 * k + 1];
  }
  for (j= 0; j < INPUT_PORTS; ++j) {
    GLuint max_depth = 0;
    for (k= 0; k < INPUT_PORTS; ++k) {
      if (depth[k] > max_depth) {
        max_depth = depth[k];
        draw_order[j] = k;
      }
    }
    depth[draw_order[j]]= 0;
  }
}

static GLuint select_object(int x, int y)
{
  GLuint hitlist[HITLIST_SIZE];
  GLint hits, viewport[4];

  glSelectBuffer(HITLIST_SIZE, hitlist);
  glRenderMode(GL_SELECT);
  glGetIntegerv(GL_VIEWPORT, viewport);
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluPickMatrix(x, actual_height - y, 3, 3, viewport);
  gluPerspective (64.0, 1.5, zNear, zFar);
  draw_scene();

  /* Now do the hard part */

  hits= glRenderMode(GL_RENDER);
  if (hits == 0)
    return 0;
  else
    /* hack */
    return hitlist[4 * (hits - 1) + 3];
}

static gboolean expose_event (GtkWidget *widget, GdkEventExpose *event,
                              gpointer data)
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);

  if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext))
    return FALSE;

  order_objects();
  int k;
  for (k= 0; k < INPUT_PORTS; ++k) {
    float level= jack_level[k];
    jack_level[k] = 0.0;
    if (level < draw_level[k]) level = draw_level[k] * 0.8 + level * 0.2;
    if (level <= 1.0e-20F) level = 0.0;
    draw_level[k] = level > 1.0 ? 1.0 : level;
  }

  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_BLEND);
  glEnable(GL_LINE_SMOOTH);

  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluPerspective (64.0, 1.5, zNear, zFar);

  /* actually create the geometry */
  draw_scene();

  glFlush();

  /* Swap buffers */
  if (gdk_gl_drawable_is_double_buffered (gldrawable))
    gdk_gl_drawable_swap_buffers (gldrawable);
  else
    glFlush ();

  gdk_gl_drawable_gl_end (gldrawable);

  return TRUE;
}

static gboolean timeout (GtkWidget *widget)
{
  if (spin) {
    sphi+= 1.0;
  }
  gtk_widget_queue_draw (widget);

  return TRUE;
}

static void unrealize (GtkWidget *widget, gpointer data)
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);

  if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext))
    return;

  gdk_gl_drawable_gl_end (gldrawable);
}

static gboolean motion_notify_event (GtkWidget *widget, GdkEventMotion *event,
                                     gpointer data)
{
  if (selected < BASE_PORT_NAME) {
    if (event->state & GDK_BUTTON1_MASK) {
      sphi += (float)(event->x - lastX) / 4.0;
      stheta += (float)(lastY - event->y) / 4.0;
    }
  } else {
    int xoff = event->x - lastX, yoff = event->y - lastY;
    if (event->state & GDK_BUTTON1_MASK) {
      GLuint input= selected - BASE_PORT_NAME;
      float pan_off = cos(sphi * DEGREES) * xoff - sin(sphi * DEGREES) * yoff;
      float vol_off = sin(sphi * DEGREES) * xoff + cos(sphi * DEGREES) * yoff;

      jack_pan[input] += pan_off / MOUSE_VELOCITY_FACTOR;
      if (jack_pan[input] < 0.0) jack_pan[input] = 0.0;
      if (jack_pan[input] > 1.0) jack_pan[input] = 1.0;
      jack_vol[input] += vol_off / MOUSE_VELOCITY_FACTOR;
      if (jack_vol[input] < 0.0) jack_vol[input] = 0.0;
      if (jack_vol[input] > 1.0) jack_vol[input] = 1.0;
    }
  }

  lastX = event->x;
  lastY = event->y;

  return FALSE;
}

static gboolean scroll_event (GtkWidget *widget, GdkEventScroll *event,
                              gpointer data)
{
  if (event->direction == GDK_SCROLL_UP) {
    if (zoom_distance >= 5.0) zoom_distance -= 0.5;
    return TRUE;
  } else if (event->direction == GDK_SCROLL_DOWN) {
    if (zoom_distance <= 15.0) zoom_distance += 0.5;
    return TRUE;
  }

  return FALSE;
}

static gboolean button_press_event (GtkWidget *widget, GdkEventButton *event,
                                      gpointer data)
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);

  redraw_rate = TIMEOUT_FAST; /* while button is down */
  timeout_speed(widget, redraw_rate);
  gtk_widget_queue_draw (widget);

  if (event->button == 1) {
      lastX = event->x;
      lastY = event->y;
      gdk_gl_drawable_gl_begin (gldrawable, glcontext);
      selected = select_object(lastX, lastY);
      gdk_gl_drawable_gl_end (gldrawable);

      return TRUE;
  }

  if (event->button == 2) {
      lastX = event->x;
      lastY = event->y;
      return TRUE;
  }

  return FALSE;
}

/* For popup menu. */
static gboolean
button_press_event_popup_menu (GtkWidget      *widget,
                               GdkEventButton *event,
                               gpointer        data)
{
  if (event->button == 3) {
    gtk_menu_popup (GTK_MENU (widget), NULL, NULL, NULL, NULL,
                    event->button, event->time);
    return TRUE;
  }

  return FALSE;
}

static gboolean button_release_event (GtkWidget *widget, GdkEventButton *event,
                                    gpointer data)
{
  /* FIXME - what if press A, press B, release A ... */

  redraw_rate = spin ? TIMEOUT_FAST : TIMEOUT_SLOW;
  timeout_speed(widget, redraw_rate);
  gtk_widget_queue_draw (widget);

  return TRUE;
}

static gboolean key_press_event (GtkWidget *widget, GdkEventKey *event,
                                 gpointer data)
{
  switch (event->keyval) {

    case GDK_Left:
      if (selected >= BASE_PORT_NAME) {
        GLuint input= selected - BASE_PORT_NAME;

        jack_pan[input] -= 0.01;
        if (jack_pan[input] < 0.0) jack_pan[input] = 0.0;
        if (jack_pan[input] > 1.0) jack_pan[input] = 1.0;
      }
      break;

    case GDK_Right:
      if (selected >= BASE_PORT_NAME) {
        GLuint input= selected - BASE_PORT_NAME;

        jack_pan[input] += 0.01;
        if (jack_pan[input] < 0.0) jack_pan[input] = 0.0;
        if (jack_pan[input] > 1.0) jack_pan[input] = 1.0;
      }
      break;

    case GDK_Up:
      if (selected >= BASE_PORT_NAME) {
        GLuint input= selected - BASE_PORT_NAME;

        jack_vol[input] += 0.01;
        if (jack_vol[input] < 0.0) jack_vol[input] = 0.0;
        if (jack_vol[input] > 1.0) jack_vol[input] = 1.0;
      }
      break;

    case GDK_Down:
      if (selected >= BASE_PORT_NAME) {
        GLuint input= selected - BASE_PORT_NAME;

        jack_vol[input] -= 0.01;
        if (jack_vol[input] < 0.0) jack_vol[input] = 0.0;
        if (jack_vol[input] > 1.0) jack_vol[input] = 1.0;
      }
      break;

    case GDK_d:
      toggle_spin (widget);
      break;

    case GDK_Escape:
      gtk_main_quit ();
      break;

    default:
      return FALSE;
  }

  return TRUE;
}

static guint timeout_id = 0;

static void
timeout_speed (GtkWidget *widget, guint32 interval)
{
  if (timeout_id != 0) {
    gtk_timeout_remove(timeout_id);
  }
  timeout_id = gtk_timeout_add (interval, (GtkFunction) timeout, widget);
}

static void
timeout_none (GtkWidget *widget)
{
  if (timeout_id != 0) {
    gtk_timeout_remove(timeout_id);
    timeout_id = 0;
  }
}

static gboolean map_event (GtkWidget *widget, GdkEvent *event, gpointer data)
{
  timeout_speed (widget, redraw_rate);

  return TRUE;
}

static gboolean unmap_event (GtkWidget *widget, GdkEvent *event,
                             gpointer data)
{
  timeout_none (widget);

  return TRUE;
}

static gboolean
visibility_notify_event (GtkWidget          *widget,
                         GdkEventVisibility *event,
                         gpointer            data)
{
  if (event->state == GDK_VISIBILITY_FULLY_OBSCURED) {
    timeout_none (widget);
  } else {
    timeout_speed (widget, redraw_rate);
  }

  return TRUE;
}


static void toggle_spin (GtkWidget *widget)
{
  spin = !spin;

  redraw_rate = spin ? TIMEOUT_FAST : redraw_rate;
  timeout_speed (widget, redraw_rate);
}


static GtkWidget *create_popup_menu (GtkWidget *drawing_area)
{
  GtkWidget *menu;
  GtkWidget *menu_item;

  menu = gtk_menu_new ();

  /* Toggle animation */
  menu_item = gtk_menu_item_new_with_label ("Toggle Animation");
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
  g_signal_connect_swapped (G_OBJECT (menu_item), "activate",
                            G_CALLBACK (toggle_spin), drawing_area);
  gtk_widget_show (menu_item);

  /* Quit */
  menu_item = gtk_menu_item_new_with_label ("Quit");
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
  g_signal_connect (G_OBJECT (menu_item), "activate",
                    G_CALLBACK (gtk_main_quit), NULL);
  gtk_widget_show (menu_item);
        
  return menu;
}

static GtkWidget *create_window (GdkGLConfig *glconfig)
{
  GtkWidget *window;
  GtkWidget *vbox;
  GtkWidget *drawing_area;
  GtkWidget *menu;
  GtkWidget *status;

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window), WINDOW_TITLE);

  /* Get automatically redrawn if any of their children changed allocation. */
  gtk_container_set_reallocate_redraws (GTK_CONTAINER (window), TRUE);

  /* Connect signal handlers to the window */
  g_signal_connect (G_OBJECT (window), "delete_event",
                    G_CALLBACK (gtk_main_quit), NULL);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (window), vbox);
  gtk_widget_show (vbox);

  /* Drawing area to draw OpenGL scene */

  drawing_area = gtk_drawing_area_new ();
  gtk_widget_set_size_request (drawing_area, DEFAULT_WIDTH, DEFAULT_HEIGHT);

  /* Set OpenGL-capability to the widget */
  gtk_widget_set_gl_capability (drawing_area,
                                glconfig,
                                NULL,
                                TRUE,
                                GDK_GL_RGBA_TYPE);

  gtk_widget_add_events (drawing_area,
                         GDK_BUTTON1_MOTION_MASK    |
                         GDK_BUTTON2_MOTION_MASK    |
                         GDK_BUTTON3_MOTION_MASK    |
                         GDK_BUTTON_PRESS_MASK      |
                         GDK_BUTTON_RELEASE_MASK    |
                         GDK_SCROLL_MASK            |
                         GDK_VISIBILITY_NOTIFY_MASK);

  /* Connect signal handlers to the drawing area */
  g_signal_connect_after (G_OBJECT (drawing_area), "realize",
                          G_CALLBACK (realize), NULL);
  g_signal_connect (G_OBJECT (drawing_area), "configure_event",
                    G_CALLBACK (configure_event), NULL);
  g_signal_connect (G_OBJECT (drawing_area), "expose_event",
                    G_CALLBACK (expose_event), NULL);
  g_signal_connect (G_OBJECT (drawing_area), "unrealize",
                    G_CALLBACK (unrealize), NULL);

  g_signal_connect (G_OBJECT (drawing_area), "motion_notify_event",
                    G_CALLBACK (motion_notify_event), NULL);
  g_signal_connect (G_OBJECT (drawing_area), "button_press_event",
                    G_CALLBACK (button_press_event), NULL);
  g_signal_connect (G_OBJECT (drawing_area), "scroll_event",
                    G_CALLBACK (scroll_event), NULL);
  g_signal_connect (G_OBJECT (drawing_area), "button_release_event",
                    G_CALLBACK (button_release_event), NULL);

  /* key_press_event handler for top-level window */
  g_signal_connect_swapped (G_OBJECT (window), "key_press_event",
                            G_CALLBACK (key_press_event), drawing_area);

  /* Don't waste CPU when window isn't visible */
  g_signal_connect (G_OBJECT (drawing_area), "map_event",
                    G_CALLBACK (map_event), NULL);
  g_signal_connect (G_OBJECT (drawing_area), "unmap_event",
                    G_CALLBACK (unmap_event), NULL);
  g_signal_connect (G_OBJECT (drawing_area), "visibility_notify_event",
                    G_CALLBACK (visibility_notify_event), NULL);

  gtk_box_pack_start (GTK_BOX (vbox), drawing_area, TRUE, TRUE, 0);

  gtk_widget_show (drawing_area);

  /* Popup menu */
  menu = create_popup_menu (drawing_area);
  g_signal_connect_swapped (G_OBJECT (drawing_area), "button_press_event",
                            G_CALLBACK (button_press_event_popup_menu), menu);

  status = gtk_statusbar_new();

  gtk_box_pack_start (GTK_BOX (vbox), status, FALSE, FALSE, 0);
  gtk_widget_show (status);

  return window;
}

/* Some utility functions, more or less verbatim from gtkglext */

static GdkGLConfig *
configure_gl (void)
{
  GdkGLConfig *glconfig;

  /* Try double-buffered visual */
  glconfig = gdk_gl_config_new_by_mode (GDK_GL_MODE_RGB    |
                                        GDK_GL_MODE_DEPTH  |
                                        GDK_GL_MODE_DOUBLE);

  /* ... or a single-bufferd equivalent */
  if (glconfig == NULL) {
    glconfig = gdk_gl_config_new_by_mode (GDK_GL_MODE_RGB | GDK_GL_MODE_DEPTH);
  }

  if (glconfig == NULL) {
    /* TODO handle this with a GTK+ dialog */
    g_print ("ERROR: No appropriate OpenGL-capable visual found.\n");
    exit (1);
  }

  return glconfig;
}


int main (int argc, char *argv[])
{
  GtkWidget *window;
  GdkGLConfig *glconfig;

  jack_handle = open_jack();
  if (jack_activate(jack_handle)) {
    fprintf(stderr, "cannot activate client"); /* FIXME show dialog */
    exit(1);
  }

  /* Initialize everything */
  gtk_init (&argc, &argv);
  gtk_gl_init (&argc, &argv);
  glconfig = configure_gl ();

  /* Create and show the application window. */
  window = create_window (glconfig);
  gtk_widget_show (window);

  gtk_main ();

  return 0;
}
